#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wayland-client.h>
#include <wayland-client-protocol.h>
#include "xdg-shell-client-protocol.h"
#include "xdg-decoration-client-protocol.h"
#include <sys/mman.h>
#include <unistd.h>
#include <stdbool.h>


struct client_state {
    /* Globals */
    struct wl_display *wl_display;
    struct wl_registry *wl_registry;
    struct wl_compositor *wl_compositor;
    struct wl_shm *wl_shm;
    struct xdg_wm_base *xdg_wm_base;
    struct zxdg_decoration_manager_v1 *zxdg_decoration_manager_v1;
    /* Objects */
    struct wl_buffer *wl_buffer;
    struct wl_surface *wl_surface;
    struct xdg_surface *xdg_surface;
    struct xdg_toplevel *xdg_toplevel;
    /* State */
    float offset;
    uint32_t last_frame;
    int width, height;
    bool closed;
};


/*
 * Create a new, unique, anonymous file of the given size, and
 * return the file descriptor for it. The file descriptor is set
 * CLOEXEC. The file is immediately suitable for mmap()'ing
 * the given size at offset zero.

 * The file is suitable for buffer sharing between processes by
 * transmitting the file descriptor over Unix sockets using the
 * SCM_RIGHTS methods.
 */
int
os_create_anonymous_file(off_t size)
{
    int fd;
    fd = memfd_create("wayland-shm", MFD_CLOEXEC);

    if (fd < 0)
        return -1;

    if (ftruncate(fd, size) < 0) {
        close(fd);
        return -1;
    }

    return fd;
}

static void
wl_buffer_release(void *data, struct wl_buffer *wl_buffer)
{
    /* Sent by the compositor when it's no longer using this buffer */
    wl_buffer_destroy(wl_buffer);
}

static const struct wl_buffer_listener wl_buffer_listener = {
        .release = wl_buffer_release,
};

static struct wl_buffer *
draw_frame(struct client_state *state) {
    int width = state->width, height = state->height;
    int stride = width * 4; // 4 bytes per pixel
    int size = stride * height;

    int fd = os_create_anonymous_file(size);
    if (fd < 0) {
        fprintf(stderr, "creating a buffer file for %d B failed: %m\n",
                size);
        exit(1);
    }

    uint32_t *data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (data == MAP_FAILED) {
        fprintf(stderr, "mmap failed: %m\n");
        close(fd);
        exit(1);
    }

    struct wl_shm_pool *pool = wl_shm_create_pool(state->wl_shm, fd, size);
    struct wl_buffer *buffer = wl_shm_pool_create_buffer(pool, 0,
            width, height, stride, WL_SHM_FORMAT_XRGB8888);
    wl_shm_pool_destroy(pool);
    close(fd);

    //fprintf(stderr, "Painting pixels\n");
//    for (int n =0; n < WIDTH*HEIGHT; n++) {
//        *data++ = 0xffff;
//    }
    /* Draw checkerboxed background */
    int offset = (int)state->offset % 8;
    for (int y = 0; y < height; ++y) {
        for (int x = 0; x < width; ++x) {
            if (((x+offset) + (y+offset) / 8 * 8) % 16 < 8)
                data[y * width + x] = 0xFF666666;
            else
                data[y * width + x] = 0xFFEEEEEE;
        }
    }

    munmap(data, size);
    wl_buffer_add_listener(buffer, &wl_buffer_listener, NULL);
    return buffer;
}

static void
xdg_toplevel_configure(void *data,
               struct xdg_toplevel *xdg_toplevel, int32_t width, int32_t height,
               struct wl_array *states)
{
    struct client_state *state = data;

    if (width == 0 || height == 0) {
        /* Compositor is deferring to us */
        return;
    }
    state->width = width;
    state->height = height;
}

static void
xdg_toplevel_close(void *data, struct xdg_toplevel *toplevel)
{
    struct client_state *state = data;
    state->closed = true;
    fprintf(stderr, "xdg_toplevel_close\n");
}

static const struct xdg_toplevel_listener xdg_toplevel_listener = {
        .configure = xdg_toplevel_configure,
        .close = xdg_toplevel_close,
};

static void
handle_configure(void *data, struct xdg_surface *xdg_surface, uint32_t serial)
{
    struct client_state *state = data;
    xdg_surface_ack_configure(xdg_surface, serial);
    fprintf(stderr, "xdg_surface configured\n");

    struct wl_buffer *buffer = draw_frame(state);
    wl_surface_attach(state->wl_surface, buffer, 0, 0);
    wl_surface_damage(state->wl_surface, 0, 0, INT32_MAX, INT32_MAX);
    wl_surface_commit(state->wl_surface);
}

static const struct xdg_surface_listener xdg_surface_listener = {
    .configure = handle_configure,
};

static void
shm_format(void *data, struct wl_shm *wl_shm, uint32_t format)
{
    //struct display *d = data;

    //        d->formats |= (1 << format);
    fprintf(stderr, "Format %d\n", format);
}

struct wl_shm_listener shm_listener = {
        shm_format
};

static void
xdg_wm_base_ping(void *data, struct xdg_wm_base *xdg_wm_base, uint32_t serial)
{
    xdg_wm_base_pong(xdg_wm_base, serial);
}

static const struct xdg_wm_base_listener xdg_wm_base_listener = {
        .ping = xdg_wm_base_ping,
};

static void
global_registry_handler(void *data, struct wl_registry *registry, uint32_t id,
               const char *interface, uint32_t version)
{
    struct client_state *state = data;

    if (strcmp(interface, wl_compositor_interface.name) == 0) {
        state->wl_compositor = wl_registry_bind(registry, id,
                                                &wl_compositor_interface, 1);

    } else if (strcmp(interface, wl_shm_interface.name) == 0) {
        state->wl_shm = wl_registry_bind(registry, id, &wl_shm_interface, 1);
        wl_shm_add_listener(state->wl_shm, &shm_listener, state);

    } else if (strcmp(interface, xdg_wm_base_interface.name) == 0) {
        state->xdg_wm_base = wl_registry_bind(registry, id,
                                              &xdg_wm_base_interface, 1);
        xdg_wm_base_add_listener(state->xdg_wm_base, &xdg_wm_base_listener, state);
    } else if (strcmp(interface, zxdg_decoration_manager_v1_interface.name) == 0) {
        state->zxdg_decoration_manager_v1 = wl_registry_bind(registry, id,
                                &zxdg_decoration_manager_v1_interface, 1);
    }
}

static void
global_registry_remover(void *data, struct wl_registry *registry, uint32_t id)
{
    printf("Got a registry losing event for %d\n", id);
}

static const struct wl_registry_listener registry_listener = {
    global_registry_handler,
    global_registry_remover
};

static const struct wl_callback_listener wl_surface_frame_listener;

static void
wl_surface_frame_done(void *data, struct wl_callback *cb, uint32_t time)
{
    // Destroy this callback
    wl_callback_destroy(cb);

    // Request another frame
    struct client_state *state = data;
    cb = wl_surface_frame(state->wl_surface);
    wl_callback_add_listener(cb, &wl_surface_frame_listener, state);

    if (state->last_frame != 0) {
        int elapsed = time - state->last_frame;
        state->offset += elapsed / 1000.0 * 24;
    }

    struct wl_buffer *buffer = draw_frame(state);
    wl_surface_attach(state->wl_surface, buffer, 0, 0);
    wl_surface_damage(state->wl_surface, 0, 0, INT32_MAX, INT32_MAX);
    wl_surface_commit(state->wl_surface);

    state->last_frame = time;
}

static const struct wl_callback_listener wl_surface_frame_listener = {
        .done = wl_surface_frame_done,
};

int main(int argc, char **argv) {

    struct client_state state = { 0 };
    state.width = 640;
    state.height = 480;
    state.closed = false;

    state.wl_display = wl_display_connect(NULL);
    if (state.wl_display == NULL) {
        fprintf(stderr, "Can't connect to display\n");
        exit(1);
    }
    printf("connected to display\n");

    state.wl_registry = wl_display_get_registry(state.wl_display);
    wl_registry_add_listener(state.wl_registry, &registry_listener, &state);

    wl_display_dispatch(state.wl_display);
    wl_display_roundtrip(state.wl_display);

    /* By now we should have wl_compositor, wl_shm and xdg_wm_base from global_registry */

    if (state.wl_compositor == NULL) {
        fprintf(stderr, "Can't find compositor\n");
        exit(1);
    } else {
        fprintf(stderr, "Found compositor\n");
    }

    state.wl_surface = wl_compositor_create_surface(state.wl_compositor);
    if (state.wl_surface == NULL) {
        fprintf(stderr, "Can't create surface\n");
        exit(1);
    } else {
        fprintf(stderr, "Created surface\n");
    }
    struct wl_callback *cb = wl_surface_frame(state.wl_surface);
    wl_callback_add_listener(cb, &wl_surface_frame_listener, &state);

    if (state.xdg_wm_base == NULL) {
        fprintf(stderr, "no wm_base\n");
        exit(1);
    }
    state.xdg_surface = xdg_wm_base_get_xdg_surface(state.xdg_wm_base, state.wl_surface);
    if (state.xdg_surface == NULL) {
        fprintf(stderr, "no xdg_surface\n");
        exit(1);
    } else {
        fprintf(stderr, "Got xdg_surface\n");
    }
    xdg_surface_add_listener(state.xdg_surface, &xdg_surface_listener, &state);

    state.xdg_toplevel = xdg_surface_get_toplevel(state.xdg_surface);
    if (state.xdg_toplevel == NULL) {
        fprintf(stderr, "no xdg_toplevel\n");
        exit(1);
    } else {
        fprintf(stderr, "got xdg_toplevel\n");
    }

    if (state.zxdg_decoration_manager_v1 != NULL){
        struct zxdg_toplevel_decoration_v1 *decoration = \
                    zxdg_decoration_manager_v1_get_toplevel_decoration(state.zxdg_decoration_manager_v1,
                                                           state.xdg_toplevel);
        zxdg_toplevel_decoration_v1_set_mode(decoration, ZXDG_TOPLEVEL_DECORATION_V1_MODE_SERVER_SIDE);
        fprintf(stderr, "Using serverside decorations");
    } else {
        fprintf(stderr, "Serverside decoration not available");
    }


    xdg_toplevel_add_listener(state.xdg_toplevel, &xdg_toplevel_listener, &state);
    xdg_toplevel_set_title(state.xdg_toplevel, "A surface");
    wl_surface_commit(state.wl_surface);


    while (wl_display_dispatch(state.wl_display) && state.closed == false) {
        ;
    }

    wl_display_disconnect(state.wl_display);
    printf("disconnected from display\n");

    exit(0);
}
